#include <Arduino.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <SparkFunTMP102.h>
#include <Adafruit_MPR121.h>

enum Pinout {
  DRIVE1 = 0,
  DRIVE2 = 1,
  LED_W = 2,
  LED_Y = 3,
  LED_B = 4,
  LED_G = 5,
  LED_R = 6,
  MATRIX_IN1 = 7,
  SBU1 = 8,
  SBU2 = 9,
  VBUS_DET = 10,
  BOOT2 = 11,
  I2C_OLED_SDA = 12,
  I2C_OLED_SCL = 13,
  WS2811 = 14,
  TX_EN = 15,
  TX = 16,
  RX = 17,
  MATRIX_IN2 = 18,
  REED3 = 19,
  REED4 = 20,
  REED2 = 21,
  REED1 = 22,
  MATRIX_IN3 = 23,
  DIGOUT1 = 24,
  DIGOUT2 = 25,
  ANALOG_IN1 = 26,
  CURRENT2 = 27,
  CURRENT1 = 28,
  MEASURE_VCC = 29,
};

const int SCREEN_ADDRESS = 0x3C;
Adafruit_SSD1306 display(128, 64, &Wire, -1);
bool hasDisplay;

#include "logo.h"

TMP102 sensor0, sensor1;

Adafruit_MPR121 cap0, cap1;
bool capAvailable[2];

void setup() {
  rp2040.enableDoubleResetBootloader();

  pinMode(DRIVE1, OUTPUT); digitalWrite(DRIVE1, LOW);
  pinMode(DRIVE2, OUTPUT); digitalWrite(DRIVE2, LOW);
  pinMode(2, OUTPUT);
  pinMode(3, OUTPUT);
  pinMode(4, OUTPUT);
  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);

  pinMode(BOOT2, INPUT_PULLUP);
  pinMode(WS2811, INPUT_PULLUP);
  pinMode(REED1, INPUT_PULLUP);
  pinMode(REED2, INPUT_PULLUP);
  pinMode(REED3, INPUT_PULLUP);
  pinMode(REED4, INPUT_PULLUP);
  pinMode(MATRIX_IN1, INPUT_PULLUP);
  pinMode(MATRIX_IN3, INPUT_PULLUP);

  Serial1.setTX(TX);  // UART0
  Serial1.setRX(RX);
  Wire.setSDA(I2C_OLED_SDA);  // I2C0
  Wire.setSCL(I2C_OLED_SCL);

  Serial.begin(9600);  // USB serial, baudrate is ignored
  Wire.begin();
  Wire.setClock(400000);

  display.setRotation(2);
  hasDisplay = display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS);
  if (!hasDisplay) {
    Serial.println(F("SSD1306 allocation failed"));
  } else {
    display.clearDisplay();
    display.drawBitmap(0, 0, logo_data, logo_width, logo_height, SSD1306_WHITE);
    display.display();
  }

  if (sensor0.begin(0x48)) {
    //0-3: 0:0.25Hz, 1:1Hz, 2:4Hz, 3:8Hz
    sensor0.setConversionRate(2);
    sensor0.setExtendedMode(1);
    sensor0.wakeup();
  }
  if (sensor1.begin(0x49)) {
    //0-3: 0:0.25Hz, 1:1Hz, 2:4Hz, 3:8Hz
    sensor1.setConversionRate(2);
    sensor1.setExtendedMode(1);
    sensor1.wakeup();
  }

  capAvailable[0] = cap0.begin(0x5a);
  capAvailable[1] = cap1.begin(0x5c);
  for (int i=0;i<2;i++) {
    if (!capAvailable[i])
      continue;
    auto& cap = (i == 0 ? cap0 : cap1);

    // disable it, again
    cap.writeRegister(MPR121_ECR, 0x0);

    // auto-config code (mostly) copied from Adafruit library
    cap.writeRegister(MPR121_AUTOCONFIG0, 0x0B);

    // correct values for Vdd = 3.3V
    cap.writeRegister(MPR121_UPLIMIT, 200);     // ((Vdd - 0.7)/Vdd) * 256
    cap.writeRegister(MPR121_TARGETLIMIT, 180); // UPLIMIT * 0.9
    cap.writeRegister(MPR121_LOWLIMIT, 130);    // UPLIMIT * 0.65

    // enable X electrodes and start MPR121
    // enable virtual proximity electrode, as well
    cap.writeRegister(MPR121_ECR, 0x8c | 0x30);

  }
}

void i2c_scan() {
  // copied from here and adapter: https://playground.arduino.cc/Main/I2cScanner/
  for (byte address = 0; address <= 127; address++ ) {
    if (address == 0) {
      Serial.print("00: ");
    } else if ((address % 16) == 0) {
      Serial.write("\r\n");
      Serial.print(address, HEX);
      Serial.write(":");
    }

    // The i2c_scanner uses the return value of
    // the Write.endTransmisstion to see if
    // a device did acknowledge to the address.
    Wire.beginTransmission(address);
    byte error = Wire.endTransmission();

    if (error == 0)
    {
      Serial.write(" ");
      if (address < 16)
        Serial.write("0");
      Serial.print(address, HEX);
    } else if (error==4) {
      Serial.print(" !!");
    } else {
      Serial.print(" --");
    }
  }
  Serial.println();
}

void displayPrintHexFixed(uint32_t x, int width) {
  while (width > 1 && (x >> (4*(width-1))) == 0) {
    display.write('0');
    width--;
  }
  display.print(x, HEX);
}

void loop() {
  static int led_timer = 0;
  static int led_state = 2;
  if (led_timer < 500)
    led_timer++;
  else {
    led_timer = 0;
    led_state++;
    if (led_state == 8)
      led_state = 2;
    if (led_state > 2)
      digitalWrite(led_state-1, 0);
    if (led_state <= 6)
      digitalWrite(led_state, 1);
  }

  static int print_state_timer = 0;
  if (led_timer < 500)
    print_state_timer++;
  else {
    print_state_timer = 0;
    Serial.write("x: ");
    Serial.write('0' + digitalRead(VBUS_DET));
    Serial.write('0' + digitalRead(BOOT2));
    Serial.write('0' + digitalRead(WS2811));
    Serial.write('0' + digitalRead(REED1));
    Serial.write('0' + digitalRead(REED2));
    Serial.write('0' + digitalRead(REED3));
    Serial.write('0' + digitalRead(REED4));
    Serial.write(", f_cpu=");
    Serial.print(rp2040.f_cpu());
    Serial.write(", id=");
    Serial.print(rp2040.cpuid());
    Serial.write(", cycles=");
    Serial.print(rp2040.getCycleCount64());
    Serial.write(", rand=");
    Serial.print(rp2040.hwrand32());
    Serial.write("\r\n");
  }

  static int i2c_scan_timer = 0;
  if (i2c_scan_timer < 5000)
    i2c_scan_timer++;
  else {
    i2c_scan_timer = 0;
    i2c_scan();
  }

  bool buttons[4];
  if (1) {
    // Button Matrix:
    // 1: SDA, IN1
    // 2: SDA, IN3
    // 3: SCL, IN1
    // 4: SCL, IN4
    Wire.end();
    pinMode(I2C_OLED_SDA, OUTPUT);
    pinMode(I2C_OLED_SCL, INPUT_PULLUP);
    delayMicroseconds(1);
    buttons[0] = !digitalRead(MATRIX_IN1);
    buttons[1] = !digitalRead(MATRIX_IN3);
    pinMode(I2C_OLED_SDA, INPUT_PULLUP);
    pinMode(I2C_OLED_SCL, OUTPUT);
    delayMicroseconds(1);
    buttons[2] = !digitalRead(MATRIX_IN1);
    buttons[3] = !digitalRead(MATRIX_IN3);
    pinMode(I2C_OLED_SDA, INPUT_PULLUP);
    pinMode(I2C_OLED_SCL, INPUT_PULLUP);
    Wire.begin();
  }

  static int displayMode = 0;
  int displayModePrev = displayMode;
  static bool prevButton4 = 0;
  if (!prevButton4 && buttons[3]) {
    displayMode = (displayMode + 1) % 5;
  }
  prevButton4 = buttons[3];

  if (rp2040.getCycleCount64() > 2 * 133*1000*1000 && displayMode == 0) {
    if (displayModePrev != 0)
      display.drawBitmap(0, 0, logo_data, logo_width, logo_height, SSD1306_WHITE);
    display.fillRect(0, 0, 67, 64, SSD1306_BLACK);

    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(0, 1);
    display.print(F("ADDR: "));
    display.println(!digitalRead(BOOT2) ? 'X' : '_');

    display.print(F("SW:   "));
    for (auto pressed : buttons)
      display.print(pressed ? 'X' : '_');
    display.println();

    display.print(F("Reed: "));
    display.print(digitalRead(REED1) ? '^' : 'v');
    display.print(digitalRead(REED2) ? '^' : 'v');
    display.print(digitalRead(REED3) ? '^' : 'v');
    display.print(digitalRead(REED4) ? '^' : 'v');
    display.println();

    auto temp0 = sensor0.readTempC();
    auto temp1 = sensor1.readTempC();
    display.print(F("Temp: "));
    display.println(temp0, 2);
    display.print(F("      "));
    display.println(temp1, 2);

    uint16_t touched0 = capAvailable[0] ? cap0.touched() : 0;
    uint16_t touched1 = capAvailable[1] ? cap1.touched() : 0;
    display.print(F("T:"));
    if (capAvailable[0])
      displayPrintHexFixed(touched0, 4);
    else
      display.write("----");
    if (capAvailable[1])
      displayPrintHexFixed(touched1, 4);
    else
      display.write("----");
    display.println();

    display.display();
  } else if (displayMode >= 1 && displayMode <= 3) {
    int first = (displayMode-1) * 12, last = first + 11;
    display.fillRect(0, 0, 128, 64, SSD1306_BLACK);
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(0, 1);
    display.print(F("Touch: "));
    display.print(first);
    display.print("..");
    display.println(last);

    for (int channel=first; channel<=last && channel < 32; channel++) {
      auto& cap = channel < 16 ? cap0 : cap1;
      auto b = cap.baselineData(channel % 16);
      auto v = cap.filteredData(channel % 16);
      displayPrintHexFixed(b, 4);
      display.write(',');
      displayPrintHexFixed(v, 4);
      if ((channel%2) == 1)
        display.println();
      else
        display.write("  ");
    }
    display.display();
  } else if (displayMode == 4) {
    display.fillRect(0, 0, 128, 64, SSD1306_BLACK);
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(80, 1);
    display.println(F("Touch:"));

    static const uint8_t channels[] = {
      // T1..T8
      0x09, 0x08, 0x07, 0x0a, 0x05, 0x02, 0x01, 0x00,
      0x1a, 0x14, 0x11, 0x16, 0x15, 0x13, 0x12,
      0xff,
      // R1..R3, north
      0x04, 0x03, 0x06,
      0xff,
      // R1..R3, south
      0x19, 0x17, 0x18,
      0xff,
      // Screw2..4
      0x0b, 0x1b, 0x10,
      0xff,
      // Proximity north and south
      0x0c, 0x1c,
      0xff,
    };

    int y = 6;
    for (auto channel : channels) {
      if (channel == 0xff) {
        display.drawFastHLine(64, y++, 64, SSD1306_WHITE);
        continue;
      }
      auto& cap = channel < 16 ? cap0 : cap1;
      auto b = cap.baselineData(channel % 16);
      auto v = cap.filteredData(channel % 16);
      display.drawFastHLine(0, y++, b/16, SSD1306_WHITE);
      display.drawFastHLine(0, y++, v/16, SSD1306_WHITE);
    }
    display.display();
  }

  delay(1);
}